@@ -1,4 +1,4 @@
-# Copyright (C) 2010-2012 Torsten Schoenfeld <kaffeetisch@gmx.de>
+# Copyright (C) 2010-2013 Torsten Schoenfeld <kaffeetisch@gmx.de>
 #
 # This library is free software; you can redistribute it and/or modify it under
 # the terms of the GNU Library General Public License as published by the Free
@@ -20,7 +20,7 @@ use strict;
 use warnings;
 use Glib;
 
-our $VERSION = '0.013';
+our $VERSION = '0.015';
 
 use Carp;
 $Carp::Internal{(__PACKAGE__)}++;
@@ -28,10 +28,9 @@ $Carp::Internal{(__PACKAGE__)}++;
 require XSLoader;
 XSLoader::load(__PACKAGE__, $VERSION);
 
-my %FORBIDDEN_SUB_NAMES = map { $_ => 1 } qw/AUTOLOAD CLONE DESTROY BEGIN
-                                             UNITCHECK CHECK INIT END/;
 my @OBJECT_PACKAGES_WITH_VFUNCS;
-
+our %_FORBIDDEN_SUB_NAMES = map { $_ => 1 } qw/AUTOLOAD CLONE DESTROY BEGIN
+                                               UNITCHECK CHECK INIT END/;
 our %_BASENAME_TO_PACKAGE;
 our %_REBLESSERS;
 
@@ -81,6 +80,9 @@ sub setup {
   my %handle_sentinel_boolean_for = exists $params{handle_sentinel_boolean_for}
     ? map { $_ => 1 } @{$params{handle_sentinel_boolean_for}}
     : ();
+  my @use_generic_signal_marshaller_for = exists $params{use_generic_signal_marshaller_for}
+    ? @{$params{use_generic_signal_marshaller_for}}
+    : ();
 
   if (exists $params{reblessers}) {
     $_REBLESSERS{$_} = $params{reblessers}->{$_}
@@ -152,6 +154,25 @@ sub setup {
     }
   }
 
+  # Monkey-patch Glib with a generic constructor for boxed types.  Glib cannot
+  # provide this on its own because it does not know how big the struct of a
+  # boxed type is.  FIXME: This sort of violates encapsulation.
+  {
+    if (! defined &{Glib::Boxed::new}) {
+      *{Glib::Boxed::new} = sub {
+        my ($class, @rest) = @_;
+        my $boxed = Glib::Object::Introspection->_construct_boxed ($class);
+        my $fields = 1 == @rest ? $rest[0] : { @rest };
+        foreach my $field (keys %$fields) {
+          if ($boxed->can ($field)) {
+            $boxed->$field ($fields->{$field});
+          }
+        }
+        return $boxed;
+      }
+    }
+  }
+
   foreach my $name (@{$interfaces}) {
     my $adder_name = $package . '::' . $name . '::_ADD_INTERFACE';
     *{$adder_name} = sub {
@@ -160,55 +181,79 @@ sub setup {
     };
   }
 
-  foreach my $object_name (keys %{$objects_with_vfuncs}) {
+  foreach my $object_name (@{$objects_with_vfuncs}) {
     my $object_package = $package . '::' . $object_name;
     my $installer_name = $object_package . '::_INSTALL_OVERRIDES';
     *{$installer_name} = sub {
       my ($target_package) = @_;
-
-      # For each vfunc in our ancestry that has an implementation, add a
-      # wrapper sub to our immediate parent.
-      my @non_perl_parent_packages =
-        __PACKAGE__->_find_non_perl_parents($basename, $object_name,
-                                            $target_package);
-      my $first_parent = $non_perl_parent_packages[0];
-      foreach my $parent_package (@non_perl_parent_packages) {
-        my @vfuncs = __PACKAGE__->_find_vfuncs_with_implementation(
-                       $parent_package, $first_parent);
-        VFUNC:
-        foreach my $vfunc_names (@vfuncs) {
-          my ($vfunc_name, $perl_vfunc_name) = @{$vfunc_names};
-          if (exists $FORBIDDEN_SUB_NAMES{$perl_vfunc_name}) {
-            $perl_vfunc_name .= '_VFUNC';
-          }
-          my $full_perl_vfunc_name =
-            $first_parent . '::' . $perl_vfunc_name;
-          if (defined &{$full_perl_vfunc_name}) {
-            next VFUNC;
-          }
-          *{$full_perl_vfunc_name} = sub {
-            __PACKAGE__->_invoke_fallback_vfunc($parent_package,
-                                                $vfunc_name,
-                                                $first_parent,
-                                                @_);
-          }
-        }
-      }
-
       # Delay hooking up the vfuncs until INIT so that we can see whether the
-      # package defines the relevant subs or not.
+      # package defines the relevant subs or not.  FIXME: Shouldn't we only do
+      # the delay dance if ${^GLOBAL_PHASE} eq 'START'?
       push @OBJECT_PACKAGES_WITH_VFUNCS,
            [$basename, $object_name, $target_package];
     };
   }
+
+  foreach my $packaged_signal (@use_generic_signal_marshaller_for) {
+    __PACKAGE__->_use_generic_signal_marshaller_for (@$packaged_signal);
+  }
 }
 
 sub INIT {
   no strict qw(refs);
+
+  # Hook up the implemented vfuncs first.
   foreach my $target (@OBJECT_PACKAGES_WITH_VFUNCS) {
     my ($basename, $object_name, $target_package) = @{$target};
     __PACKAGE__->_install_overrides($basename, $object_name, $target_package);
   }
+
+  # And then, for each vfunc in our ancestry that has an implementation, add a
+  # wrapper sub to our immediate parent.  We delay this step until after all
+  # Perl overrides are in place because otherwise, the override code would see
+  # the fallback vfuncs (via gv_fetchmethod) we are about to set up, and it
+  # would mistake them for an actual implementation.  This would then lead it
+  # to put Perl callbacks into the vfunc slots regardless of whether the Perl
+  # class in question actually provides implementations.
+  my %implementer_packages_seen;
+  foreach my $target (@OBJECT_PACKAGES_WITH_VFUNCS) {
+    my ($basename, $object_name, $target_package) = @{$target};
+    my @non_perl_parent_packages =
+      __PACKAGE__->_find_non_perl_parents($basename, $object_name,
+                                          $target_package);
+
+    # For each non-Perl parent, look at all the vfuncs it and its parents
+    # provide.  For each vfunc which has an implementation in the parent
+    # (i.e. the corresponding struct pointer is not NULL), install a fallback
+    # sub which invokes the vfunc implementation.  This assumes that
+    # @non_perl_parent_packages contains the parents in "ancestorial" order,
+    # i.e. the first entry must be the immediate parent.
+    IMPLEMENTER: for (my $i = 0; $i < @non_perl_parent_packages; $i++) {
+      my $implementer_package = $non_perl_parent_packages[$i];
+      next IMPLEMENTER if $implementer_packages_seen{$implementer_package}++;
+      for (my $j = $i; $j < @non_perl_parent_packages; $j++) {
+        my $provider_package = $non_perl_parent_packages[$j];
+        my @vfuncs = __PACKAGE__->_find_vfuncs_with_implementation(
+                       $provider_package, $implementer_package);
+        VFUNC: foreach my $vfunc_name (@vfuncs) {
+          my $perl_vfunc_name = uc $vfunc_name;
+          if (exists $_FORBIDDEN_SUB_NAMES{$perl_vfunc_name}) {
+            $perl_vfunc_name .= '_VFUNC';
+          }
+          my $full_perl_vfunc_name =
+            $implementer_package . '::' . $perl_vfunc_name;
+          next VFUNC if defined &{$full_perl_vfunc_name};
+          *{$full_perl_vfunc_name} = sub {
+            __PACKAGE__->_invoke_fallback_vfunc($provider_package,
+                                                $vfunc_name,
+                                                $implementer_package,
+                                                @_);
+          }
+        }
+      }
+    }
+  }
+
   @OBJECT_PACKAGES_WITH_VFUNCS = ();
 }
 
@@ -341,6 +386,14 @@ be returned, and otherwise an empty list will be returned.
 The function names refer to those after name corrections.  Functions occuring
 in C<handle_sentinel_boolean_for> may also occur in C<class_static_methods>.
 
+=item use_generic_signal_marshaller_for => [ [package1, signal1, [arg_converter1]], ... ]
+
+Use an introspection-based generic signal marshaller for the signal C<signal1>
+of type C<package1>.  If given, use the code reference C<arg_converter1> to
+convert the arguments that are passed to the signal handler.  In contrast to
+L<Glib>'s normal signal marshaller, the generic signal marshaller supports,
+among other things, pointer arrays and out arguments.
+
 =item reblessers => { package => \&reblesser, ... }
 
 Tells G:O:I to invoke I<reblesser> whenever a Perl object is created for an
@@ -417,6 +470,18 @@ wrapper for a function that expects a GValue, do this:
   # substitute $real_value where you'd use $value
   ...
 
+=head2 Handling extendable enumerations
+
+If you need to handle extendable enumerations for which more than the
+pre-defined values might be valid, then use C<<
+Glib::Object::Introspection->convert_enum_to_sv >> and C<<
+Glib::Object::Introspection->convert_sv_to_enum >>.  They will raise an
+exception on unknown values; catching it then allows you to implement fallback
+behavior.
+
+  Glib::Object::Introspection->convert_enum_to_sv (package, enum_value)
+  Glib::Object::Introspection->convert_sv_to_enum (package, sv)
+
 =head1 SEE ALSO
 
 =over
